/*++++++++++++++
.IDENTIFICATION ipname.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix
.KEYWORDS       
.VERSION  1.0   25-Apr-1995
.VERSION  1.1   09-Oct-2005: Accept list in input
.VERSION  1.2   06-Sep-2010: -vp (vpar file)
.COMMENTS       From example on Solaris
---------------*/

#define VERSION "1.2"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#define is_comment(c)	((c)=='#') || ((c)=='"') || ((c)=='\\')

static char usage[] = "\
Usage: ipname [-s] [-num | IP]" ;

static char help[] = "\
    -s: silently ignore non-found IP numbers\n\
   -vp: standard input is of 'vpar' type (logs of vizier)\n\
  -num: number indicated the word position of IP number (e.g. -7 for xferlog)\n\
   IP : Numeric IP number\n\
" ;

typedef struct aNode { 
   struct aNode *next;
   u_long ip ;
   char name[4];	/* Actual size differs */
} NODE ;

static NODE *list[256] ;

static char verbop = 0;
static char opt_s = 0;

static NODE *find_ip (char *host)
/*++++++++++++++++
.PURPOSE  Find the name corresponding to an IP number
.RETURNS  The corresponding name
.REMARKS  Added if not yet existing
-----------------*/
{
  NODE *n, *p ;
  char *h, *hnum, *hnam ;
  struct hostent *hp;
  /* struct in_addr inaddr ; */
  u_long addr ;	
  int i, x ;

    hnum = hnam = (char *)0 ;
    for (h=host; isgraph(*h); h++) ;

    /* Is it an IP name or number ? */
    for (h=host, i=0; isdigit(*h) || (*h == '.'); h++) {
        if (*h == '.') i++; 
    }
    if ((i==3) && ((!*h) || isspace(*h) || (*h == '(')))  /* Yes, IP number */
	hnum = host ;
    else {
	hnam = host;
        while (isgraph(*h)) h++ ; while (isspace(*h)) h++ ;
	if (*h == '(') h++ ;
	hnum = h ;
	while (isdigit(*h) || (*h == '.')) h++;
    }
    if (!hnum) return((NODE *)0) ;

    /* Convert the IP number, h = end of hnum string */
    x = *h; *h = 0; 
    addr = inet_addr(hnum); 
    *h = x ;
    if ((int)(addr) == -1) {
	if (opt_s) ;
        else fprintf(stderr, "#***Bad IP: %s\n", hnum);
	return ((NODE *)0) ;
    }

    /* Look: is the IP number already saved ? */
    i = ((addr>>24) + (addr>>8) + (addr&0xff)) & 0xff ;
    n = p = list[i] ;
    while (n) {
        p = n ;
	if (n->ip == addr) break ;
	n = n -> next ;
    }
    if (n) 		/* It's there !! */
        return(n) ;

    /* Not yet Found... Is the host name available ? */
    if (!hnam) {
        hp = gethostbyaddr((char *)&addr, sizeof (addr), AF_INET);
        if (hp) hnam = hp->h_name ;
	else {
	    x = *h; *h = 0;
	    if (opt_s) ; else fprintf(stderr, 
	      "++++host information for %08lX (%s) not found\n", 
	      addr, hnum);
	    *h = x ;
	    hnam = hnum ;
	}
    }

    /* Allocate the new corespondance */
    for (h=hnam; isgraph(*h); h++) ;
    x = h-hnam ;
    n = (NODE *)malloc(sizeof(NODE) + x - 3) ;
    n->ip = addr ;
    memcpy(n->name, hnam, x) ;
    n->name[x] = 0 ;
    n->next = (NODE *)0 ;
    if (p) p->next = n ;
    else list[i] = n ;

    return(n) ;
}


int main(int argc, char **argv) 
{
  int found = 0;
  int word =  1;
  char *xline = 0;
  char *arg, *p;
  char line[4096];
  int i, len, len1=0, nd, comline, cont;
  NODE *n;

    while (--argc>0) {
	arg = *++argv;
	if (*arg == '-') switch(arg[1]) {
	  case 's':
	    opt_s = 1;
	    continue;
	  case 'v':
	    if (strncmp(arg, "-ver", 4) == 0) {
		printf("aserver (CDS) Version %s\n", VERSION);
		exit(0);
	    }
	    if (arg[2] == 'p') {	/* -vpar */
		word = 6;
		xline = "----"; 	/* Paragraph line to look at */
		continue;
	    }
	    verbop = 1;
	    continue;
	  case 'h':	/* Help */
	    printf("%s", help);
	    exit(0);
	  default:
	    if (isdigit(arg[1])) {
		word = atoi(arg+1);	/* Offset in word */
		continue;
	    }
	    fprintf(stderr, "****%s\n", usage);
	    exit(1);
	}

	/* A single argument: we assume it's an IP number */
	found++;
	if ((n = find_ip(arg))) printf("%s\n", n->name);
    }

    if (found>0) exit(0);

    /* Take names from stdin */

    cont = 0;
    while (fgets(line, sizeof(line), stdin)) {
	len = strlen(line) -1;
	if (cont) {		/* A continuation */
	    printf("%s", line);
	    if (line[len] == '\n') cont = 0;
	    continue;
	}
	found++;

	cont = line[len] != '\n' ;

	/* Is it a good line (not a omment line) */
	if (xline) {
	    if (len1==0) {
	        for (i=0; xline[i]==line[i]; i++) ;
		comline = xline[i];
	    }
	    else comline = 1;
	}
	else comline = is_comment(line[0]);

	/* len1: without trailing spaces */
	for (len1=len; (len1>0) && isspace(line[len1]); len1--) ;
	if (len1 < 7) comline = 1;

	if (comline) { 	/* Not a good line */
	    printf("%s", line);
	    continue;
	}

	/* Locate the word */
	for (p=line; isspace(*p); p++) ;
	for (i=word; --i>0; ) {
	    while (isgraph(*p)) p++;
	    while (isspace(*p)) p++;
	}

	/* Does the work look like an IP number ? */
	if (isdigit(*p)) {
	    for (i=nd=0; isgraph(p[i]); i++) {
		if (p[i] == '.') nd++; 
		else if (!isdigit(p[i])) break;
	    }
	    if ((nd==3)&&isspace(p[i])) ;  /* OK */
	    else {
		/*
		nd = p[i]; p[i] = 0;
		fprintf(stderr, "#+++Bad IP number '%s' (line#%d)\n", p, found);
		p[i] = nd; */
		printf("%s", line);
		continue;
	    }
	}
	else if (xline) {	/* Change only the IP numbers */
	    printf("%s", line);
	    continue;
	}

	if ((n = find_ip(p))) {
	    fwrite(line, 1, p-line, stdout);
	    printf("%s", n->name);
	    while (isgraph(*p)) p++;  /* Skip the IP number */
	    printf("%s", p);
	}
	else {
	    printf("%s", line);
	}
    }
    exit (0);
}
